import DB as db
import operators as op
from data import QRTable
books, books_authors, authors, events = [QRTable()] * 4

# todo add 'as' syntax: 'select count(*) as cnt
# todo when dict format used, manage same names in different tables

def main():
    DB = db.DB('postgres', 'qrook_db', 'kurush', 'pondoxo', format='dict')
    DB.create_data(__name__, in_module=True)
    print(books, books.id)
    data = DB.select(books).where(original_publication_year=2000, language_code='eng').\
        where(id=op.In(470, 490, 485)).all()
    print(data[:10])

    data = books.select('count(*)').group_by(books.original_publication_year).all()
    print(data[:10])

    data = DB.select(books, books.id).where('id < 10').order_by(books.id, desc=True).\
        limit(3).offset(2).all()
    print('limit & offset', data)

    data = books.select(authors.id, books.id)\
        .join(books_authors, 'books_authors.book_id = books.id')\
        .join(authors, op.Eq(books_authors.author_id, authors.id)).all()
    print(data[:10])

    data = books.select(books.id).where(id=1).where(bool='or', id=2).all()
    print('with or:', data)

    #ok = DB.delete(events, auto_commit=True).exec()
    #print(ok)

    from datetime import datetime
    t = datetime.now().time()
    d = datetime.now().date()
    ok = DB.update(events, auto_commit=True).set(time=t).where(id=6).exec()
    print(ok)

    #ok = DB.insert(events, events.time, auto_commit=True).values([t]).exec()
    #ok = DB.insert(events, events.date, events.time, auto_commit=True).values([d, t]).exec()
    query = events.insert(events.date, events.time, auto_commit=True).values([[d, t], [None, t]]).returning('*')
    data = query.all()
    print(data)

    data = DB.exec('select * from get_book_authors(1) as f(id int, name varchar)').config_fields('id', 'name').all()
    print(data)


if __name__ == '__main__':
    main()


from request import *
from PostgresConnector import *
from error_handlers import *
from data import *
import sys
from data_formatter import defaultDataFormatter

# todo add exec raw query
# todo add support for nested queries
# todo add select(books.id) parsing -> extract table from ONLY datafield

@log_class(log_error)
class DB:
    @log_error

    # format: one of 'list', 'dict'
    def __init__(self, connector_type, *conn_args, format=None, **conn_kwargs):
        self.meta = dict()
        if connector_type == 'postgres':
            self.meta['connector'] = PostgresConnector(*conn_args, **conn_kwargs)
        else:
            raise Exception('unknown connector type')

        if format is not None:
            defaultDataFormatter.format = format

    # source is object with __dict__ field or a module name (with 'in_module' flag up)
    def create_data(self, source=None, in_module=False):
        tables = self.meta['connector'].table_info()
        t = dict()

        for name, field in tables.items():
            t[name] = QRTable(name, field, self)

        self.__dict__.update(t)
        if source:
            if in_module:
                source = sys.modules[source]
            source.__dict__.update(t)

    def commit(self):
        self.meta['connector'].commit()

    def exec(self, raw_query):
        #logger.warning('UNSAFE: executing raw query: %s', raw_query)
        return QRequest(self.meta['connector'], request=raw_query)

    def select(self, table: QRTable, *args):
        identifiers, literals, used_fields = [], [], []
        if len(args) == 0:
            own_args = list(table.meta['fields'].values())
        else:
            own_args = args

        fields = ''
        for arg in own_args:
            if isinstance(arg, QRField):
                fields += '{}.{},'
                identifiers.extend([arg.table_name, arg.name])
                used_fields.append(arg.name)
            else:
                #logger.warning('UNSAFE: executing raw select from table %s with args: %s', table, args)
                fields += arg + ','
                used_fields.append(arg)
        fields = fields[:-1]

        request = 'select ' + fields + ' from {}'
        table_name = table.meta['table_name']
        identifiers += [table_name]

        return QRSelect(self.meta['connector'], table, request,
                        identifiers, literals, used_fields=used_fields)

    def delete(self, table: QRTable, auto_commit=False):
        identifiers, literals = [], []

        request = 'delete from {}'
        table_name = table.meta['table_name']
        identifiers += [table_name]

        return QRWhere(self.meta['connector'], table, request, identifiers,
                       literals, auto_commit)

    def update(self, table: QRTable, auto_commit=False):
        identifiers, literals = [], []

        request = 'update {}'
        table_name = table.meta['table_name']
        identifiers += [table_name]

        return QRUpdate(self.meta['connector'], table, request, identifiers,
                             literals, auto_commit)

    def insert(self, table: QRTable, *args, auto_commit=False):
        identifiers, literals = [], []
        if len(args) == 0:
            fields = ''
        else:
            fields = ''
            for arg in args:
                if isinstance(arg, QRField):
                    fields += '{},'
                    identifiers.extend([arg.name])
                else:
                    #logger.warning('UNSAFE: executing raw select from table %s with args: %s', table, args)
                    fields += arg + ','
            fields = fields[:-1]
            fields = '(' + fields + ')'

        request = 'insert into {} ' + fields + ' values '
        table_name = table.meta['table_name']
        identifiers = [table_name] + identifiers

        return QRInsert(self.meta['connector'], table, request, identifiers, literals, auto_commit)



from error_handlers import *


@log_class(log_error)
class QRTable:
    # fields = {'a':'int', 'b':'varchar'}
    def __init__(self, table_name=None, fields=None, DB=None):
        self.meta = dict()
        self.meta['table_name'] = table_name
        self.meta['fields'] = {}
        self.DB = DB

        if fields is None: return
        for name, value_type in fields:
            f = QRField(name, value_type, self)
            self.meta['fields'][name] = f
            self.__dict__[name] = f

    def __str__(self):
        if self.meta['table_name'] is None:
            return '<Empty QRTable>'
        return '<QRTable ' + self.meta['table_name'] + '>'

    def select(self, *args, **kwargs):
        if self.DB is None:
            raise Exception('no DB instance set for table to make queries to')
        return self.DB.select(self, *args, **kwargs)

    def update(self, *args, **kwargs):
        if self.DB is None:
            raise Exception('no DB instance set for table to make queries to')
        return self.DB.update(self, *args, **kwargs)

    def insert(self, *args, **kwargs):
        if self.DB is None:
            raise Exception('no DB instance set for table to make queries to')
        return self.DB.insert(self, *args, **kwargs)

    def delete(self, *args, **kwargs):
        if self.DB is None:
            raise Exception('no DB instance set for table to make queries to')
        return self.DB.delete(self, *args, **kwargs)


class QRField:
    def __init__(self, name, value_type, table: QRTable):
        self.name = name
        self.type = value_type
        self.table_name = table.meta['table_name']
        self.table = table

    def __str__(self):
        if self.name is None:
            return '<Empty QRField>'
        return '<QRField ' + self.name + ' of ' + self.table_name + '>'



import psycopg2
import psycopg2.sql as sql

from Connector import Connector
from error_handlers import log_error, retry_log_error
from qrlogging import logger


class PostgresConnector(Connector):
    def __init__(self, db, user, password, host='localhost', port=5432):
        super().__init__(db, user, password, host, port)
        self.__connect()

    def __del__(self):
        if self.__dict__.get('conn'):
            self.conn.close()

    @retry_log_error()
    def __connect(self):
        self.conn = psycopg2.connect(dbname=self.db, user=self.user,
                                     password=self.password, host=self.host, port=self.port)
        self.cursor = self.conn.cursor()

    @log_error
    def exec(self, request: str, identifiers: list = None, literals: list = None, result='all'):
        super().exec(request, identifiers, literals, result)
        request = sql.SQL(request)
        if identifiers:
            identifiers = [sql.Identifier(x) for x in identifiers]
            request = request.format(*identifiers)

        logger.warning('POSTGRES EXECUTE: %s with literals %s', request.as_string(self.cursor), literals)
        self.cursor.execute(request, literals)
        return self.extract_result(result)

    def extract_result(self, result):
        if result == 'all':
            return self.cursor.fetchall()
        elif result == 'one':
            return self.cursor.fetchone()
        return None

    @log_error
    def table_info(self):
        request = "select table_name, column_name, data_type  " \
                  "from information_schema.columns " \
                  "where table_schema = 'public'"
        data = self.exec(request, result='all')
        info = {}
        for d in data:
            if not info.get(d[0]):
                info[d[0]] = []
            info[d[0]].append((d[1], d[2]))
        return info

    @log_error
    def commit(self):
        self.conn.commit()



from error_handlers import *


class Connector:
    connected = False
    db = None
    user = None
    password = None
    host = None
    port = None

    def __init__(self, db, user, password, host, port):
        self.db = db
        self.user = user
        self.password = password
        self.host = host
        self.port = port

    def __del__(self):
        pass

    # for 'all': [(1, 2, 3), ...] or []
    # for 'one': (1,2,3) or None
    def exec(self, request_str: str, identifiers: list = None, literals: list = None, result='all'):

        pass

    # {'books':[('id', 'integer'), ('date', 'date'), ...], ...}
    def table_info(self):
        pass

    def commit(self):
        pass



from data import QRField


class QROperator:
    def __init__(self, op, data):
        self.data = data
        self.op = op

    def condition(self, disable_full_name=False):
        if disable_full_name:
            return '{}' + self.op + '%s', [self.data]
        else:
            return '{}.{}' + self.op + '%s', [self.data]

class Between(QROperator):
    def __init__(self, a, b):
        super().__init__('<>', [a, b])

    def condition(self, disable_full_name=False):
        if disable_full_name:
            return '{} between %s and %s', list(self.data)
        else:
            return '{}.{} between %s and %s', list(self.data)

class In(QROperator):
    def __init__(self, *args):
        super().__init__('<>', args)

    def condition(self, disable_full_name=False):
        likes = ','.join(['%s'] * len(self.data))
        if disable_full_name:
            return '{} in(' + likes + ')', list(self.data)
        else:
            return '{}.{} in(' + likes + ')', list(self.data)

class Eq(QROperator):
    def __init__(self, arg1, arg2=None):
        self.arg1 = arg1
        self.arg2 = arg2
        self.duos = arg2 is not None
        super().__init__('=', arg1)

    def condition(self, disable_full_name=False):
        if not self.duos:
            return super().condition(disable_full_name)
        else:
            if disable_full_name:
                return '{} = {}', []
            else:
                return '{}.{} = {}.{}', []

class GT(QROperator):
    def __init__(self, data):
        super().__init__('>', data)

class GE(QROperator):
    def __init__(self, data):
        super().__init__('>=', data)

class LT(QROperator):
    def __init__(self, data):
        super().__init__('<', data)

class LE(QROperator):
    def __init__(self, data):
        super().__init__('<=', data)

class NE(QROperator):
    def __init__(self, data):
        super().__init__('<>', data)

class Like(QROperator):
    def __init__(self, data):
        super().__init__(' like ', data)



# class like - update condition() method


import logging
# todo use configs for filename and logger name


_log_format = "%(asctime)s [%(levelname)s]: [%(app)s, %(name)s] - " \
         "(%(filename)s).%(funcName)s(%(lineno)d) - %(message)s"


class CustomFormatter(logging.Formatter):
    """Logging Formatter to add colors and count warning / errors"""
    grey = "\033[30m"
    yellow = "\033[33m"
    red = "\033[31m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"

    FORMATS = {
        logging.DEBUG: grey + _log_format + reset,
        logging.INFO: grey + _log_format + reset,
        logging.WARNING: yellow + _log_format + reset,
        logging.ERROR: red + _log_format + reset,
        logging.CRITICAL: bold_red + _log_format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def get_file_handler():
    file_handler = logging.FileHandler("qrook_app.log")
    file_handler.setLevel(logging.WARNING)
    file_handler.setFormatter(logging.Formatter(_log_format))
    return file_handler


def get_stream_handler():
    sh = logging.StreamHandler()
    sh.setLevel(logging.INFO)
    sh.setFormatter(logging.Formatter(_log_format))
    sh.setFormatter(CustomFormatter())
    return sh


def default_logger():
    '''fast setup
    logging.basicConfig(filename="qrook_app.log", level=logging.INFO,
                        format = "%(asctime)s - [%(levelname)s] - %(app)s - %(name)s -"\
                         "(%(filename)s).%(funcName)s(%(lineno)d) - %(message)s")'''

    logger = logging.getLogger("default")
    logger.addHandler(get_file_handler())
    logger.addHandler(get_stream_handler())
    logger = logging.LoggerAdapter(logger, {"app": "qrook"})
    return logger


logger = default_logger()
